/*
 * direct_cp.c
 * 
 * Copyright (c) 2011-2013, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
 */

#define PAGE_SIZE				4096
#define DEFAULT_BUFFER_SIZE		(2 * 1024 * 1024)
#define WAL_FILE_NAME_LENGTH	24
#define MESSAGE_MAX_LENGTH		256
#define PROGRAM_NAME			"direct_cp"
#define PROGRAM_VERSION			"1.0.0"

#include "postgres.h"
#include "access/xlog_internal.h"

#include <unistd.h>
#include <sys/stat.h>
#include <regex.h>
#include <getopt.h>

static void call_cp(char *ifpath, char *ofpath);
static int reg_check(char *strcheck, char *strregex);
static void show_usage(void);
static void show_version(void);
static void die(const char *fmt, ...)
__attribute__((format(PG_PRINTF_ATTRIBUTE, 1, 2)));
static bool parse_int32(const char *value, int32 *result);

/*
 * function name : call_cp
 *
 * 1st argument : input file name
 * 2nd argument : output file name
 * 
 * return : nothing
 * 
 * Call "cp" command using system() function.
 * Input file is copied to output file.
 */
static void
call_cp(char *ifpath, char *ofpath)
{
	char	command[MAXPGPATH];

	snprintf(command, sizeof(command), "cp -f %s %s", ifpath, ofpath);

	if (system(command) != 0)
		die("cp command failed.");
}

/*
 * function name : reg_check
 *
 * 1st argument : strcheck
 * 2nd argument : strregex
 *
 * return : 0 or 1
 *
 * check string(strcheck) include regular expression string(strregex).
 * strcheck include strregex       : result = 0
 * strcheck don't include strregex : result = 1
 */
static int
reg_check(char *strcheck, char *strregex)
{
	regex_t	preg;
	int		result;
	int		errnum;
	char	errbuf[256];

	if ((errnum = regcomp(&preg, strregex, REG_EXTENDED|REG_NOSUB)))
	{
		regerror(errnum, &preg, errbuf, 100);
		die("regex compile failed. : %s", errbuf);
	}

	if ((regexec(&preg, strcheck, 0, NULL, 0) != REG_NOMATCH))
		result = 0;
	else
		result = 1;

	regfree(&preg);

	return result;
}

int
main(int argc, char **argv)
{
	char	 ifpath[MAXPGPATH];	/* input file path */
	char	 ofpath[MAXPGPATH];	/* output file path */
	char	*ifname;
	int		 ifd = 0;	/* input file directive */
	int		 ofd = 0;	/* output file directive */
	int		 ret, cnt;	/* return, counter */
	void	*buf;		/* buffer */
	int		 bufsize = DEFAULT_BUFFER_SIZE;
	int		 opt;
	int		 opt_index;
	struct stat	ifinfo, ofinfo;

	static struct option options[] = {
		{ "buffer-size", required_argument, NULL, 'b' },
		{ "version", no_argument, NULL, 'V' },
		{ "help", no_argument, NULL, '?' },
		{ NULL, 0, NULL, 0 }
	};

	/* check option arguments */
	opterr = 0;
	while ((opt = getopt_long(argc, argv, "V?b:", options, &opt_index)) != -1)
	{
		switch (opt)
		{
			/* command option of '-b' is setting buffered area size */
			case 'b':
				if (!parse_int32(optarg, &bufsize))
					die("buffer size convert failed.");
				break;
			/* command option of '-V' or '--version' is showing version information */
			case 'V':
				show_version();
				exit(0);
			case '?':
				/* command option of '-?' or '--help' is showing help information */
				if (strcmp(argv[optind - 1], "-?") == 0 ||
					strcmp(argv[optind - 1], "--help") == 0)
				{
					show_usage();
					exit(0);
				}
				else
				{
					/* unknown option reported by getopt */
					die("[%s] is undefined.", argv[optind - 1]);
				}
				break;
			default:
				break;
		}
	}

	/* as a number of argument is more or less than 2 files, finish this program */
	argc -= optind;
	if (argc < 2)
		die("leak out designation of input file or output file.");
	else if (argc >= 3)
		die("too many arguments are designated.");

	StrNCpy(ifpath, argv[optind], sizeof(ifpath));
	StrNCpy(ofpath, argv[optind + 1], sizeof(ofpath));

	/* check buffered size what is less than PAGE_SIZE or more than XLOG_SEG_SIZE */
	if (bufsize < PAGE_SIZE || bufsize > XLOG_SEG_SIZE)
		die("buffer size[%d] must be more than page size[%d] and less than WAL file size[%d].",
			bufsize, PAGE_SIZE, XLOG_SEG_SIZE);

	/* check buffered size what is a multiple of PAGE_SIZE */
	if (bufsize % PAGE_SIZE != 0)
		die("buffer size[%d] is not a multiple of page size[%d].",
			bufsize, PAGE_SIZE);

	/* allocate buffered area */
	
	if ((ret = posix_memalign(&buf, PAGE_SIZE, bufsize)) != 0)
		die("memory allocation failed. : %s", strerror(ret));

	memset(buf, 0, bufsize);

	ifname = strrchr(ifpath, '/');
	if (ifname != NULL)
		ifname++;
	else
		ifname = ifpath;

	if (stat(ifpath, &ifinfo) != 0)
		die("cannot stat input file[%s]. : %s", ifpath, strerror(errno));

	/* the case that input file name is WAL file */
	if (reg_check(ifname, "([[:xdigit:]]{24})") == 0 &&
		strlen(ifname) == WAL_FILE_NAME_LENGTH &&
		ifinfo.st_size == XLOG_SEG_SIZE)
	{
		/* check what input and output file is a same file. */
		if (stat(ofpath, &ofinfo) == 0)
		{
			if (ifinfo.st_ino == ofinfo.st_ino)
				die("input file and output file is a same file[%s].", ofpath);

			if (S_ISDIR(ofinfo.st_mode))
			{
				join_path_components(ofpath, ofpath, ifname);

				if (stat(ofpath, &ofinfo) == 0 &&
					ifinfo.st_ino == ofinfo.st_ino)
					die("input file and output file is a same file[%s].", ofpath);
			}
		}

		/* open the input and output file on "O_DIRECT" mode */
		if ((ifd = open(ifpath, O_RDONLY|O_DIRECT)) < 0)
			die("open input file[%s] failed. : %s", ifpath, strerror(errno));

		if ((ofd = open(ofpath, O_CREAT|O_TRUNC|O_WRONLY|O_DIRECT, S_IRWXU)) < 0)
			die("open output file[%s] failed. : %s", ofpath, strerror(errno));

		/* read input file and write to output file */
		cnt = 0;
		while (1)
		{
			ret = pread(ifd, buf, bufsize, cnt * bufsize);
			if (ret == bufsize)
			{
				if (pwrite(ofd, buf, bufsize, cnt * bufsize) < 0)
				{
					remove(ofpath);
					die("write to output file[%s] failed. : %s", ofpath, strerror(errno));
				}
				cnt++;
			}
			else if (ret < 0)
			{
				remove(ofpath);
				die("read input file[%s] failed. : %s", ifpath, strerror(errno));
			}
			else
				break;
		}
	}
	/* the case that input file name is "not" 24 consecutive numbers */
	else
	{
		/* call "cp" command using system() function */
		call_cp(ifpath, ofpath);
	}

	/* close input and output file, release buffered area, and return 0 to finish this program */
	close(ifd);
	close(ofd);
	free(buf);

	return 0;
}

static void
die(const char *fmt, ...)
{
	va_list		ap;
	char		msg[MESSAGE_MAX_LENGTH];

	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);

	fprintf(stderr, "%s\n", msg);
	fflush(stderr);

	exit(1);
}

static bool
parse_int32(const char *value, int32 *result)
{
	int64	val;
	char   *endptr;

	Assert(value);

	errno = 0;
	val = strtol(value, &endptr, 0);
	if (endptr == value || *endptr)
		return false;	// illegal characters

	if (errno == ERANGE || val != (int64) ((int32) val))
		return false;	// out of range

	*result = (int32) val;

	return true;
}

static void
show_version(void)
{
	printf("%s %s\n", PROGRAM_NAME, PROGRAM_VERSION);
}

static void
show_usage(void)
{
	printf("direct_cp is direct-IO copy utility for WAL files.\n\n");
	printf("Usage:\n");
	printf("  direct_cp [OPTION]... SOURCE DEST\n");
	printf("Copy SOURCE to DEST\n\n");
	printf("Options:\n");
	printf("  -b, --buffer-size=NUM  set buffer size in Byte\n");
	printf("                         must be a multiple of %d\n", PAGE_SIZE);
	printf("                         and must be from %d to %d (default: %d)\n", PAGE_SIZE, XLOG_SEG_SIZE, DEFAULT_BUFFER_SIZE);
	printf("  -V, --version          output version information, then exit\n");
	printf("  -?, --help             show this help, then exit\n");
}
